# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Any

import flask as flsk
import imageio as mgio
from si_fi_o.session.session import session_t

import pca_b_stream.main as pcas


output_types_h = tuple[tuple[str | None, int, Any], path_t | None]


def ProcessSession(
    session: session_t, session_id: str, folder: path_t, /
) -> output_types_h:
    """"""
    # FIXME: Find a better way to guess targeted conversion
    if "to byte stream" in flsk.request.form:
        return _ToSTR(session, session_id, folder)
    else:
        return _ToARY(session, session_id, folder)


def _ToSTR(session: session_t, _: str, __: path_t, /) -> output_types_h:
    """"""
    image = mgio.imread(session["array"].path)
    if pcas.PCAIsValid(image):
        stream = pcas.PCA2BStream(image)
        length = stream.__len__()
        details = pcas.BStreamDetails(
            stream,
            should_print=False,
            should_return=True,
        )
        stream = stream.decode("ascii")
    else:
        stream = "Invalid Piecewise-constant Array"
        length = -1
        details = None

    return (stream, length, details), None


def _ToARY(session: session_t, session_id: str, folder: path_t, /) -> output_types_h:
    """"""
    stream = bytes(session["stream"], "ascii")
    stream = b"\n".join(stream.splitlines())  # To convert newlines to \n

    try:
        decoded = pcas.BStream2PCA(stream)
    except:
        decoded = None

    length = stream.__len__()
    if decoded is None:
        details = None
        path = None
    else:
        details = pcas.BStreamDetails(
            stream,
            should_print=False,
            should_return=True,
        )
        path = folder / f"decoded-{session_id}.png"
        folder.mkdir(exist_ok=True)
        mgio.imwrite(path, decoded)

    return (None, length, details), path
